import time
import unittest

import fluteline
from .basic_nodes import Producer, Consumer


class TestProducer(unittest.TestCase):

    def setUp(self):
        self.producer = Producer()
        self.producer.output = fluteline.Queue()
        self.producer.start()

    def tearDown(self):
        self.producer.stop()

    def test_producer(self):
        item = self.producer.output.get()
        self.assertEqual(item, 1)


class TestPipeline(unittest.TestCase):

    def setUp(self):
        self.producer = Producer()
        self.consumer = Consumer()
        self.consumer.output = fluteline.Queue()
        self.producer.output = self.consumer
        self.producer.start()
        self.consumer.start()

    def tearDown(self):
        self.producer.stop()
        self.consumer.stop()

    def test_pipeline(self):
        item = self.consumer.output.get()
        self.assertEqual(item, 2)


if __name__ == '__main__':
    unittest.main()



try:
    import queue  # python 3
except ImportError:
    import Queue as queue


class Queue(object):
    '''
    Thread-safe queue for nodes communication.
    '''
    def __init__(self):
        self._queue = queue.Queue()

    def empty(self):
        '''
        Return ``True`` if the queue is empty, ``False`` otherwise
        (not reliable!).
        '''
        return self._queue.empty()

    def put(self, item):
        '''
        Put an item into the queue.
        '''
        return self._queue.put(item)

    def get(self):
        '''
        Remove and return an item from the queue.
        '''
        return self._queue.get()



import abc
import threading
import warnings

from . import queue


class _TerminationMessage(object):
    '''
    Send an instance to a node input to stop it.
    '''
    pass


class Node(threading.Thread):
    '''
    An abstract base class for producers, consumers, or consumer-producers
    in a pipeline.

    Hooks
        Setup resources when starting in ``enter`` and clean after yourself
        in ``exit``.

    Start and stop
        Call ``start`` to spin up the node. Don't forget to call ``stop``.
        Otherwise the thread will stay alive.
    '''
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        super(Node, self).__init__()
        self.input = queue.Queue()
        self._stopping = False

    def consume(self, msg):
        '''
        Override if your node should react to incoming message, whether it's
        a consumer, consumer-producer, or just a producer.

        :param msg: An incoming messages to process.
        '''
        warnings.warn('A message arrived to a node with no `consume` implementation')

    def enter(self):
        '''
        Override to prepare resources.
        '''
        pass

    def exit(self):
        '''
        Override to cleanup after yourself.
        '''
        pass

    def put(self, msg):
        '''
        Send a message to this node.
        '''
        self.input.put(msg)

    def stop(self):
        '''
        Stop the node gracefully.
        '''
        self.put(_TerminationMessage())

    def run(self):
        self.enter()
        try:
            while not self._stopping:
                self._loop()
        finally:
            self.exit()

    @abc.abstractmethod
    def _loop(self):
        pass

    def _process_input(self):
        msg = self.input.get()
        if isinstance(msg, _TerminationMessage):
            self._stopping = True
        else:
            self.consume(msg)


class Producer(Node):
    '''
    Inherit this class to create a producer.
    '''
    def produce(self):
        '''
        Override to produce new messages.
        '''
        pass

    def _loop(self):
        if self.input.empty():
            self.produce()
        else:
            self._process_input()


class Consumer(Node):
    '''
    Inherit this class to create a consumer or consumer-producer.
    '''
    def _loop(self):
        self._process_input()



import unittest

import fluteline
from .basic_nodes import Producer, Consumer


class TestUtilities(unittest.TestCase):

    def test_connect(self):
        n1 = Producer()
        n2 = Consumer()
        n3 = Consumer()
        nodes = [n1, n2, n3]
        fluteline.connect(nodes)
        self.assertEqual(n1.output, n2)
        self.assertEqual(n2.output, n3)
        self.assertIsInstance(n3.output, fluteline.Queue)



from . import nodes
from . import queue


def connect(nodes):
    '''
    Connect a list of nodes.

    Connected nodes have an ``output`` member which is the following node in
    the line. The last node's ``output`` is a :class:`fluteline.Queue` for
    easy plumbing.
    '''
    for a, b in zip(nodes[:-1], nodes[1:]):
        a.output = b
    b.output = queue.Queue()


def start(nodes):
    '''
    Start multiple nodes.
    '''
    for node in nodes:
        node.start()


def stop(nodes):
    '''
    Stop multiple nodes.
    '''
    for node in nodes:
        node.stop()
